@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2000-2002  Maxim Krasnyansky <maxk@qualcomm.com>
- *  Copyright (C) 2003-2007  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2003-2011  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -33,62 +33,38 @@
 #include <stdlib.h>
 #include <string.h>
 #include <getopt.h>
+#include <sys/poll.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 
-#include <bluetooth/bluetooth.h>
-#include <bluetooth/hci.h>
-#include <bluetooth/hci_lib.h>
-
 #include <arpa/inet.h>
-#include <netinet/in.h>
 #include <netdb.h>
 
 #include "parser/parser.h"
 #include "parser/sdp.h"
 
-#if __BYTE_ORDER == __LITTLE_ENDIAN
-static inline uint64_t ntoh64(uint64_t n)
-{
-	uint64_t h;
-	uint64_t tmp = ntohl(n & 0x00000000ffffffff);
-	h = ntohl(n >> 32);
-	h |= tmp << 32;
-	return h;
-}
-#elif __BYTE_ORDER == __BIG_ENDIAN
-#define ntoh64(x) (x)
-#else
-#error "Unknown byte order"
-#endif
-#define hton64(x) ntoh64(x)
+#include "lib/hci.h"
+#include "lib/hci_lib.h"
 
-#define SNAP_LEN 	HCI_MAX_FRAME_SIZE
-#define DEFAULT_PORT	"10839";
+#define SNAP_LEN	HCI_MAX_FRAME_SIZE
+#define DEFAULT_PORT   "10839"
 
 /* Modes */
 enum {
 	PARSE,
 	READ,
 	WRITE,
-	RECEIVE,
-	SEND,
+	SERVER,
 	PPPDUMP,
 	AUDIO
 };
 
 /* Default options */
-static int  device;
 static int  snap_len = SNAP_LEN;
-static int  defpsm = 0;
-static int  defcompid = DEFAULT_COMPID;
 static int  mode = PARSE;
 static int  permcheck = 1;
-static int  noappend = 0;
-static long flags;
-static long filter;
 static char *dump_file = NULL;
 static char *pppdump_file = NULL;
 static char *audio_file = NULL;
@@ -136,7 +112,7 @@ struct pktlog_hdr {
 
 static inline int read_n(int fd, char *buf, int len)
 {
-	register int t = 0, w;
+	int t = 0, w;
 
 	while (len > 0) {
 		if ((w = read(fd, buf, len)) < 0) {
@@ -153,7 +129,7 @@ static inline int read_n(int fd, char *buf, int len)
 
 static inline int write_n(int fd, char *buf, int len)
 {
-	register int t = 0, w;
+	int t = 0, w;
 
 	while (len > 0) {
 		if ((w = write(fd, buf, len)) < 0) {
@@ -168,7 +144,7 @@ static inline int write_n(int fd, char *buf, int len)
 	return t;
 }
 
-static void process_frames(int dev, int sock, int fd, unsigned long flags)
+static int process_frames(int dev, int sock, int fd, unsigned long flags)
 {
 	struct cmsghdr *cmsg;
 	struct msghdr msg;
@@ -176,9 +152,14 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	struct hcidump_hdr *dh;
 	struct btsnoop_pkt *dp;
 	struct frame frm;
+	struct pollfd fds[2];
+	int nfds = 0;
 	char *buf, *ctrl;
 	int len, hdr_size = HCIDUMP_HDR_SIZE;
 
+	if (sock < 0)
+		return -1;
+
 	if (snap_len < SNAP_LEN)
 		snap_len = SNAP_LEN;
 
@@ -188,7 +169,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	buf = malloc(snap_len + hdr_size);
 	if (!buf) {
 		perror("Can't allocate data buffer");
-		exit(1);
+		return -1;
 	}
 
 	dh = (void *) buf;
@@ -197,8 +178,9 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 	ctrl = malloc(100);
 	if (!ctrl) {
+		free(buf);
 		perror("Can't allocate control buffer");
-		exit(1);
+		return -1;
 	}
 
 	if (dev == HCI_DEV_NONE)
@@ -206,11 +188,72 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 	else
 		printf("device: hci%d ", dev);
 
-	printf("snap_len: %d filter: 0x%lx\n", snap_len, filter);
+	printf("snap_len: %d filter: 0x%lx\n", snap_len, parser.filter);
 
 	memset(&msg, 0, sizeof(msg));
 
+	if (mode == SERVER) {
+		struct btsnoop_hdr *hdr = (void *) buf;
+
+		btsnoop_version = 1;
+		btsnoop_type = 1002;
+
+		memcpy(hdr->id, btsnoop_id, sizeof(btsnoop_id));
+		hdr->version = htonl(btsnoop_version);
+		hdr->type = htonl(btsnoop_type);
+
+		printf("btsnoop version: %d datalink type: %d\n",
+						btsnoop_version, btsnoop_type);
+
+		len = write(fd, buf, BTSNOOP_HDR_SIZE);
+		if (len < 0) {
+			perror("Can't create dump header");
+			return -1;
+		}
+
+		if (len != BTSNOOP_HDR_SIZE) {
+			fprintf(stderr, "Header size mismatch\n");
+			return -1;
+		}
+
+		fds[nfds].fd = fd;
+		fds[nfds].events = POLLIN;
+		fds[nfds].revents = 0;
+		nfds++;
+	}
+
+	fds[nfds].fd = sock;
+	fds[nfds].events = POLLIN;
+	fds[nfds].revents = 0;
+	nfds++;
+
 	while (1) {
+		int i, n = poll(fds, nfds, -1);
+		if (n <= 0)
+			continue;
+
+		for (i = 0; i < nfds; i++) {
+			if (fds[i].revents & (POLLHUP | POLLERR | POLLNVAL)) {
+				if (fds[i].fd == sock)
+					printf("device: disconnected\n");
+				else
+					printf("client: disconnect\n");
+				return 0;
+			}
+		}
+
+		if (mode == SERVER) {
+			len = recv(fd, buf, snap_len, MSG_DONTWAIT);
+			if (len == 0) {
+				printf("client: disconnect\n");
+				return 0;
+			}
+			if (len < 0 && errno != EAGAIN && errno != EINTR) {
+				perror("Connection read failure");
+				return -1;
+			}
+		}
+
 		iv.iov_base = frm.data;
 		iv.iov_len  = snap_len;
 
@@ -219,10 +262,12 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 		msg.msg_control = ctrl;
 		msg.msg_controllen = 100;
 
-		len = recvmsg(sock, &msg, 0);
+		len = recvmsg(sock, &msg, MSG_DONTWAIT);
 		if (len < 0) {
+			if (errno == EAGAIN || errno == EINTR)
+				continue;
 			perror("Receive failed");
-			exit(1);
+			return -1;
 		}
 
 		/* Process control message */
@@ -234,12 +279,15 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 		cmsg = CMSG_FIRSTHDR(&msg);
 		while (cmsg) {
+			int dir;
 			switch (cmsg->cmsg_type) {
 			case HCI_CMSG_DIR:
-				frm.in = *((int *) CMSG_DATA(cmsg));
+				memcpy(&dir, CMSG_DATA(cmsg), sizeof(int));
+				frm.in = (uint8_t) dir;
 				break;
 			case HCI_CMSG_TSTAMP:
-				frm.ts = *((struct timeval *) CMSG_DATA(cmsg));
+				memcpy(&frm.ts, CMSG_DATA(cmsg),
+						sizeof(struct timeval));
 				break;
 			}
 			cmsg = CMSG_NXTHDR(&msg, cmsg);
@@ -250,7 +298,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 		switch (mode) {
 		case WRITE:
-		case SEND:
+		case SERVER:
 			/* Save or send dump */
 			if (flags & DUMP_BTSNOOP) {
 				uint64_t ts;
@@ -273,7 +321,7 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 
 			if (write_n(fd, buf, frm.data_len + hdr_size) < 0) {
 				perror("Write error");
-				exit(1);
+				return -1;
 			}
 			break;
 
@@ -283,6 +331,8 @@ static void process_frames(int dev, int sock, int fd, unsigned long flags)
 			break;
 		}
 	}
+
+	return 0;
 }
 
 static void read_dump(int fd)
@@ -404,12 +454,9 @@ static int open_file(char *file, int mode, unsigned long flags)
 	struct btsnoop_hdr *hdr = (struct btsnoop_hdr *) buf;
 	int fd, len, open_flags;
 
-	if (mode == WRITE || mode == PPPDUMP || mode == AUDIO) {
-		if (noappend || flags & DUMP_BTSNOOP)
-			open_flags = O_WRONLY | O_CREAT | O_TRUNC;
-		else
-			open_flags = O_WRONLY | O_CREAT | O_APPEND;
-	} else
+	if (mode == WRITE || mode == PPPDUMP || mode == AUDIO)
+		open_flags = O_WRONLY | O_CREAT | O_TRUNC;
+	else
 		open_flags = O_RDONLY;
 
 	fd = open(file, open_flags, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
@@ -492,19 +539,19 @@ static int open_socket(int dev, unsigned long flags)
 		dd = hci_open_dev(dev);
 		if (dd < 0) {
 			perror("Can't open device");
-			exit(1);
+			return -1;
 		}
 
 		if (hci_devinfo(dev, &di) < 0) {
 			perror("Can't get device info");
-			exit(1);
+			return -1;
 		}
 
 		opt = hci_test_bit(HCI_RAW, &di.flags);
 		if (ioctl(dd, HCISETRAW, opt) < 0) {
 			if (errno == EACCES) {
 				perror("Can't access device");
-				exit(1);
+				return -1;
 			}
 		}
 
@@ -515,19 +562,19 @@ static int open_socket(int dev, unsigned long flags)
 	sk = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_HCI);
 	if (sk < 0) {
 		perror("Can't create raw socket");
-		exit(1);
+		return -1;
 	}
 
 	opt = 1;
 	if (setsockopt(sk, SOL_HCI, HCI_DATA_DIR, &opt, sizeof(opt)) < 0) {
 		perror("Can't enable data direction info");
-		exit(1);
+		return -1;
 	}
 
 	opt = 1;
 	if (setsockopt(sk, SOL_HCI, HCI_TIME_STAMP, &opt, sizeof(opt)) < 0) {
 		perror("Can't enable time stamp");
-		exit(1);
+		return -1;
 	}
 
 	/* Setup filter */
@@ -536,165 +583,215 @@ static int open_socket(int dev, unsigned long flags)
 	hci_filter_all_events(&flt);
 	if (setsockopt(sk, SOL_HCI, HCI_FILTER, &flt, sizeof(flt)) < 0) {
 		perror("Can't set filter");
-		exit(1);
+		return -1;
 	}
 
 	/* Bind socket to the HCI device */
+	memset(&addr, 0, sizeof(addr));
 	addr.hci_family = AF_BLUETOOTH;
 	addr.hci_dev = dev;
 	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
 		printf("Can't attach to device hci%d. %s(%d)\n", 
 					dev, strerror(errno), errno);
-		exit(1);
+		return -1;
 	}
 
 	return sk;
 }
 
-static int open_connection(char *addr, char *port)
+static int create_datagram(unsigned short port)
 {
-	struct sockaddr_storage ss;
-	struct addrinfo hints, *res0, *res;
-	int sk = -1, opt = 1;
-	
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = af;
-	hints.ai_socktype = SOCK_STREAM;
-	hints.ai_protocol = IPPROTO_TCP;
-	
-	if (getaddrinfo(addr, port, &hints, &res0))
-		if(getaddrinfo(NULL, port, &hints, &res0)) {
-			perror("getaddrinfo");
-			exit(1);
-		}
-	
-	for (res = res0; res; res = res->ai_next) {
-		sk = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
-		if (sk < 0) {
-			if (res->ai_next)
-				continue;
+	struct sockaddr_in addr;
+	int sk, opt = 1;
 
-			perror("Can't create socket");
-			freeaddrinfo(res0);
-			exit(1);
-		}
+	sk = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
+	if (sk < 0)
+		return -1;
 
-		setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
-		
-		memcpy(&ss, res->ai_addr, res->ai_addrlen);
-
-		switch(ss.ss_family) {
-		case AF_INET:
-			((struct sockaddr_in *) &ss)->sin_addr.s_addr = htonl(INADDR_ANY);
-			((struct sockaddr_in *) &ss)->sin_port = 0;
-			break;
-		case AF_INET6:
-			memcpy(&((struct sockaddr_in6 *) &ss)->sin6_addr, &in6addr_any, sizeof(in6addr_any));
-			((struct sockaddr_in6 *) &ss)->sin6_port = 0;
-			break;
-		}
-
-		if (bind(sk, (struct sockaddr *) &ss, sizeof(ss)) < 0) {
-			perror("Can't bind socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
-		}
-		
-		if (connect(sk, res->ai_addr, res->ai_addrlen) < 0) {
-			perror("Can't connect socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
-		}
+	if (setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) < 0) {
+		close(sk);
+		return -1;
 	}
 
-	freeaddrinfo(res0);
+	memset(&addr, 0, sizeof(addr));
+	addr.sin_family = AF_INET;
+	addr.sin_port = htons(port);
+	addr.sin_addr.s_addr = htonl(INADDR_BROADCAST);
+
+	if (bind(sk, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		close(sk);
+		return -1;
+	}
 
 	return sk;
 }
 
+static unsigned char ping_data[] = { 'p', 'i', 'n', 'g' };
+static unsigned char pong_data[] = { 'p', 'o', 'n', 'g' };
+
+static void handle_datagram(int sk)
+{
+	struct sockaddr_in addr;
+	socklen_t addr_len = sizeof(addr);
+	unsigned char buf[64];
+	ssize_t len;
+
+	len = recvfrom(sk, buf, sizeof(buf), MSG_DONTWAIT,
+				(struct sockaddr *) &addr, &addr_len);
+
+	if (len != sizeof(ping_data))
+		return;
+
+	if (memcmp(buf, ping_data, sizeof(ping_data)) != 0)
+		return;
+
+	len = sendto(sk, pong_data, sizeof(pong_data), 0,
+				(struct sockaddr *) &addr, sizeof(addr));
+}
+
 static int wait_connection(char *addr, char *port)
 {
 	char hname[100], hport[10];
-	struct sockaddr_storage ss;
-	struct addrinfo hints, *res0, *res;
-	socklen_t len;
-	int sk = -1, nsk, opt = 1;
+	struct addrinfo *ai, *runp;
+	struct addrinfo hints;
+	struct pollfd fds[3];
+	unsigned int nfds = 0;
+	int err, opt, datagram;
 
-	memset(&hints, 0, sizeof(hints));
-	hints.ai_family = af;
+	memset(&hints, 0, sizeof (hints));
+	hints.ai_flags = AI_PASSIVE | AI_ADDRCONFIG;
 	hints.ai_socktype = SOCK_STREAM;
 	hints.ai_protocol = IPPROTO_TCP;
 
-	if (getaddrinfo(addr, port, &hints, &res0))
-		if (getaddrinfo(NULL, port, &hints, &res0)) {
-			perror("getaddrinfo");
-			exit(1);
-		}
- 	
-	for (res = res0; res; res = res->ai_next) {
-		sk = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
-		if (sk < 0) {
-			if (res->ai_next)
-				continue;
+	err = getaddrinfo(dump_addr, dump_port, &hints, &ai);
+	if (err < 0) {
+		printf("Can't get address info: %s\n", gai_strerror(err));
+		return -1;
+	}
 
+	runp = ai;
+
+	datagram = create_datagram(atoi(dump_port));
+	if (datagram < 0) {
+		printf("server: no discover protocol\n");
+	} else {
+		fds[nfds].fd = datagram;
+		fds[nfds].events = POLLIN;
+		nfds++;
+	}
+
+	while (runp != NULL && nfds < sizeof(fds) / sizeof(fds[0])) {
+		fds[nfds].fd = socket(runp->ai_family, runp->ai_socktype,
+							runp->ai_protocol);
+		if (fds[nfds].fd < 0) {
 			perror("Can't create socket");
- 			freeaddrinfo(res0);
-			exit(1);
+			return -1;
 		}
 
-		setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
+		fds[nfds].events = POLLIN;
 
-		if (bind(sk, res->ai_addr, res->ai_addrlen) < 0) {
-			if(res->ai_next) {
-				close(sk);
+		opt = 1;
+		setsockopt(fds[nfds].fd, SOL_SOCKET, SO_REUSEADDR,
+							&opt, sizeof(opt));
+
+		opt = 0;
+		setsockopt(fds[nfds].fd, SOL_SOCKET, SO_KEEPALIVE,
+							&opt, sizeof(opt));
+
+		if (bind(fds[nfds].fd, runp->ai_addr, runp->ai_addrlen) < 0) {
+			if (errno != EADDRINUSE) {
+				perror("Can't bind socket");
+				return -1;
+			}
+
+			close(fds[nfds].fd);
+		} else {
+			if (listen(fds[nfds].fd, SOMAXCONN) < 0) {
+				perror("Can't listen on socket");
+				return -1;
+			}
+
+			getnameinfo(runp->ai_addr, runp->ai_addrlen,
+							hname, sizeof(hname),
+							hport, sizeof(hport),
+							NI_NUMERICSERV);
+
+			printf("server: %s:%s snap_len: %d filter: 0x%lx\n",
+					hname, hport, snap_len, parser.filter);
+
+			nfds++;
+		}
+
+		runp = runp->ai_next;
+	}
+
+	freeaddrinfo(ai);
+
+	while (1) {
+		unsigned int i;
+		int n = poll(fds, nfds, -1);
+		if (n <= 0)
+			continue;
+
+		for (i = 0; i < nfds; i++) {
+			struct sockaddr_storage rem;
+			socklen_t remlen = sizeof(rem);
+			int sk;
+
+			if (!(fds[i].revents & POLLIN))
+				continue;
+
+			if (fds[i].fd == datagram) {
+				handle_datagram(datagram);
 				continue;
 			}
 
-			perror("Can't bind socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
- 		}
-
-		getnameinfo(res->ai_addr, res->ai_addrlen, hname, sizeof(hname),
-					hport, sizeof(hport), NI_NUMERICSERV);
-		printf("device: %s:%s snap_len: %d filter: 0x%lx\n", 
-					hname, hport, snap_len, filter);
-		if (listen(sk, 1) < 0) {
-			if (res->ai_next) {
-				close(sk);
+			sk = accept(fds[i].fd, (struct sockaddr *) &rem, &remlen);
+			if (sk < 0)
 				continue;
-			}
 
-			perror("Can't listen on socket");
-			close(sk);
-			freeaddrinfo(res0);
-			exit(1);
+			getnameinfo((struct sockaddr *) &rem, remlen,
+							hname, sizeof(hname),
+							hport, sizeof(hport),
+							NI_NUMERICSERV);
+
+			printf("client: %s:%s snap_len: %d filter: 0x%lx\n",
+					hname, hport, snap_len, parser.filter);
+
+			for (n = 0; n < (int) nfds; n++)
+				close(fds[n].fd);
+
+			return sk;
 		}
 	}
 
-	freeaddrinfo(res0);
+	return -1;
+}
 
-	len = sizeof(ss);
-	nsk = accept(sk, (struct sockaddr *) &ss, &len);
-	if (nsk < 0) {
-		perror("Can't accept new socket");
+static int run_server(int dev, char *addr, char *port, unsigned long flags)
+{
+	while (1) {
+		int dd, sk;
+
+		sk = wait_connection(addr, port);
+		if (sk < 0)
+			continue;
+
+		//fcntl(sk, F_SETFL, O_NONBLOCK);
+
+		dd = open_socket(dev, flags);
+		if (dd < 0) {
+			close(sk);
+			continue;
+		}
+
+		process_frames(dev, dd, sk, flags);
+
+		close(dd);
 		close(sk);
-		freeaddrinfo(res0);
-		exit(1);
 	}
 
-	getnameinfo((struct sockaddr *) &ss, sizeof(ss),
-					hname, sizeof(hname), NULL, 0, 0);
-
-	printf("device: %s snap_len: %d filter: 0x%lx\n",
-						hname, snap_len, filter);
-
-	close(sk);
-
-	return nsk;
+	return 0;
 }
 
 static struct {
@@ -711,6 +808,8 @@ static struct {
 	{ "cmtp",	FILT_CMTP	},
 	{ "hidp",	FILT_HIDP	},
 	{ "hcrp",	FILT_HCRP	},
+	{ "att",	FILT_ATT	},
+	{ "smp",	FILT_SMP	},
 	{ "avdtp",	FILT_AVDTP	},
 	{ "avctp",	FILT_AVCTP	},
 	{ "obex",	FILT_OBEX	},
@@ -721,8 +820,9 @@ static struct {
 	{ 0 }
 };
 
-static void parse_filter(int argc, char **argv)
+static unsigned long parse_filter(int argc, char **argv)
 {
+	unsigned long filter = 0;
 	int i,n;
 
 	for (i = 0; i < argc; i++) {
@@ -733,6 +833,8 @@ static void parse_filter(int argc, char **argv)
 			}
 		}
 	}
+
+	return filter;
 }
 
 static void usage(void)
@@ -745,8 +847,7 @@ static void usage(void)
 	"  -m, --manufacturer=compid  Default manufacturer\n"
 	"  -w, --save-dump=file       Save dump to a file\n"
 	"  -r, --read-dump=file       Read dump from a file\n"
-	"  -s, --send-dump=host       Send dump to a host\n"
-	"  -n, --recv-dump=host       Receive dump on a host\n"
+	"  -d, --wait-dump=host       Wait on a host and send\n"
 	"  -t, --ts                   Display time stamps\n"
 	"  -a, --ascii                Dump data in ascii\n"
 	"  -x, --hex                  Dump data in hex\n"
@@ -758,13 +859,11 @@ static void usage(void)
 	"  -P, --ppp=channel          Channel for PPP\n"
 	"  -D, --pppdump=file         Extract PPP traffic\n"
 	"  -A, --audio=file           Extract SCO audio data\n"
-	"  -B, --btsnoop              Use BTSnoop file format\n"
-	"  -V, --verbose              Verbose decoding\n"
 	"  -Y, --novendor             No vendor commands or events\n"
-	"  -N, --noappend             No appending to existing files\n"
 	"  -4, --ipv4                 Use IPv4 as transport\n"
 	"  -6  --ipv6                 Use IPv6 as transport\n"
 	"  -h, --help                 Give this help list\n"
+	"  -v, --version              Give version information\n"
 	"      --usage                Give a short usage message\n"
 	);
 }
@@ -776,8 +875,7 @@ static struct option main_options[] = {
 	{ "manufacturer",	1, 0, 'm' },
 	{ "save-dump",		1, 0, 'w' },
 	{ "read-dump",		1, 0, 'r' },
-	{ "send-dump",		1, 0, 's' },
-	{ "recv-dump",		1, 0, 'n' },
+	{ "wait-dump",		1, 0, 'd' },
 	{ "timestamp",		0, 0, 't' },
 	{ "ascii",		0, 0, 'a' },
 	{ "hex",		0, 0, 'x' },
@@ -789,24 +887,25 @@ static struct option main_options[] = {
 	{ "ppp",		1, 0, 'P' },
 	{ "pppdump",		1, 0, 'D' },
 	{ "audio",		1, 0, 'A' },
-	{ "btsnoop",		0, 0, 'B' },
-	{ "verbose",		0, 0, 'V' },
 	{ "novendor",		0, 0, 'Y' },
 	{ "nopermcheck",	0, 0, 'Z' },
-	{ "noappend",		0, 0, 'N' },
 	{ "ipv4",		0, 0, '4' },
 	{ "ipv6",		0, 0, '6' },
 	{ "help",		0, 0, 'h' },
+	{ "version",		0, 0, 'v' },
 	{ 0 }
 };
 
 int main(int argc, char *argv[])
 {
+	unsigned long flags = 0;
+	unsigned long filter = 0;
+	int device = 0;
+	int defpsm = 0;
+	int defcompid = DEFAULT_COMPID;
 	int opt, pppdump_fd = -1, audio_fd = -1;
 
-	printf("HCI sniffer - Bluetooth packet analyzer ver %s\n", VERSION);
-
-	while ((opt=getopt_long(argc, argv, "i:l:p:m:w:r:s:n:taxXRC:H:O:P:D:A:BVYZN46h", main_options, NULL)) != -1) {
+	while ((opt=getopt_long(argc, argv, "i:l:p:m:w:r:d:taxXRC:H:O:P:D:A:YZ46hv", main_options, NULL)) != -1) {
 		switch(opt) {
 		case 'i':
 			if (strcasecmp(optarg, "none") && strcasecmp(optarg, "system"))
@@ -837,13 +936,8 @@ int main(int argc, char *argv[])
 			dump_file = strdup(optarg);
 			break;
 
-		case 's':
-			mode = SEND;
-			dump_addr = optarg;
-			break;
-
-		case 'n':
-			mode = RECEIVE;
+		case 'd':
+			mode = SERVER;
 			dump_addr = optarg;
 			break;
 
@@ -891,14 +985,6 @@ int main(int argc, char *argv[])
 			audio_file = strdup(optarg);
 			break;
 
-		case 'B':
-			flags |= DUMP_BTSNOOP;
-			break;
-
-		case 'V':
-			flags |= DUMP_VERBOSE;
-			break;
-
 		case 'Y':
 			flags |= DUMP_NOVENDOR;
 			break;
@@ -907,10 +993,6 @@ int main(int argc, char *argv[])
 			permcheck = 0;
 			break;
 
-		case 'N':
-			noappend = 1;
-			break;
-
 		case '4':
 			af = AF_INET;
 			break;
@@ -919,6 +1001,10 @@ int main(int argc, char *argv[])
 			af = AF_INET6;
 			break;
 
+		case 'v':
+			printf("%s\n", VERSION);
+			exit(0);
+
 		case 'h':
 		default:
 			usage();
@@ -930,8 +1016,10 @@ int main(int argc, char *argv[])
 	argv += optind;
 	optind = 0;
 
+	printf("HCI sniffer - Bluetooth packet analyzer ver %s\n", VERSION);
+
 	if (argc > 0)
-		parse_filter(argc, argv);
+		filter = parse_filter(argc, argv);
 
 	/* Default settings */
 	if (!filter)
@@ -945,28 +1033,27 @@ int main(int argc, char *argv[])
 
 	switch (mode) {
 	case PARSE:
+		flags |= DUMP_VERBOSE;
 		init_parser(flags, filter, defpsm, defcompid, pppdump_fd, audio_fd);
 		process_frames(device, open_socket(device, flags), -1, flags);
 		break;
 
 	case READ:
+		flags |= DUMP_VERBOSE;
 		init_parser(flags, filter, defpsm, defcompid, pppdump_fd, audio_fd);
 		read_dump(open_file(dump_file, mode, flags));
 		break;
 
 	case WRITE:
+		flags |= DUMP_BTSNOOP;
 		process_frames(device, open_socket(device, flags),
 				open_file(dump_file, mode, flags), flags);
 		break;
 
-	case RECEIVE:
+	case SERVER:
+		flags |= DUMP_BTSNOOP;
 		init_parser(flags, filter, defpsm, defcompid, pppdump_fd, audio_fd);
-		read_dump(wait_connection(dump_addr, dump_port));
-		break;
-
-	case SEND:
-		process_frames(device, open_socket(device, flags),
-				open_connection(dump_addr, dump_port), flags);
+		run_server(device, dump_addr, dump_port, flags);
 		break;
 	}
 
